#============================================================================
#====================  Generic Procedures for systems of equations   ========
#====================  Author: B. Alpers, FH Aalen, 23.12.99         ========
#============================================================================
#============================================================================


#============================================================================
#======== Check equation for correctness  ===================================
#============================================================================

G_check_equation:=proc(equation,system,var_set)

#=== check whether equation is compatible with the correct system in a certain
#=== variable set var_set. Compatible means that system and system union {equation}
#=== provide the same solution.
#=== If so, the routine returns 0, else 1.

local correct_solution,extended_sol;

correct_solution:=solve(system,var_set);
if correct_solution = NULL then correct_solution:={};fi; # if solve can't solve
extended_sol:=solve(system union {equation}, var_set);
if extended_sol = NULL then extended_sol:={}; fi;  # if solve can't solve
	 
if G_compare_solution_sets(extended_sol, correct_solution, var_set)=1 then
	RETURN(1);
else 	RETURN(0);
fi;

end:

#============================================================================
#====================  Find error in wrong equation		     ========
#============================================================================

#=== Note: This might be extended to hint_system, thus using information on
#===       equations which have been set up correctly


G_find_error:=proc(equation, system, var_set)

#=== Return - if possible - the kind of error, the wrong part of the equation,
#=== and the correct equation
#=== The error types currently implemented are:
#=== 0 = NO_ERROR
#=== 1 = NOT_IDENTIFIED (none of the below)
#=== 2 = WRONG_SYMBOLS (set of wrong symbols)
#=== 3 = SIGN_ERROR (set of terms in the sum with wrong signs)
#=== 4 = WRONG_FACTORS (set of terms in the sum with wrong factors other than -1)
#=== 5 = WRONG_TERMS (set of wrong terms other than in 3 and 4)
#=== 6 = MISSING_TERMS (set of missing terms)


local 	NO_ERROR, NOT_IDENTIFIED, WRONG_SYMBOLS, SIGN_ERROR, WRONG_FACTORS,
	WRONG_TERMS, MISSING_TERMS, 
	setlist, correct_equation, sign_error_set, wrong_factor_term_set,
	term_error_set, missing_term_set, compare_result;

# Initialize error type variables
NO_ERROR:=0;
NOT_IDENTIFIED:=1;
WRONG_SYMBOLS:=2;
SIGN_ERROR:=3;
WRONG_FACTORS:=4;
WRONG_TERMS:=5;
MISSING_TERMS:=6;

#=== check for errors

if G_check_equation(equation, system, var_set)=0 then
   RETURN([0,{}]);
fi;


#=== first check whether the equation contains a wrong symbol

setlist:=G_compare_symbol_sets(equation, system);
if setlist[1] <> {} then 
	RETURN([WRONG_SYMBOLS, setlist[1]]);
fi;


#=== next compare equation with all correct equations

for correct_equation in system do
	compare_result:=G_compare_equations(equation, correct_equation, var_set);
	
	# extract sets
	sign_error_set:=compare_result[2];
	wrong_factor_term_set:=compare_result[3];
	term_error_set:=compare_result[5];
	missing_term_set:=compare_result[6];

	# identify specific erroneous situations and give hints
	
	# only sign errors:

	if (sign_error_set <> {}) and wrong_factor_term_set={} and
	   term_error_set = {}  and missing_term_set = {}
	then RETURN([SIGN_ERROR, sign_error_set, correct_equation]); fi;

	# only wrong numerical factors:
	# WARNING: there might be sign errors resulting from writing A=B
	# instead of B=A leading to A-B=0 instead of B-A=0. So, maybe the
	# sign_error_set should not required to be empty! Same holds for other
	# cases below

	if sign_error_set = {} and wrong_factor_term_set <> {} and
	   term_error_set = {}  and missing_term_set = {}
	then RETURN([WRONG_FACTORS, wrong_factor_term_set, correct_equation]); fi;

	# just one wrong term (and maybe the correct one missing)
	# if there is more than one term wrong the two equation are
	# probably totally different although this depends on the number
	# of terms present in the equation; could be changed to condition
	# nops(term_error_set)/nops(lhs(equation)-rhs(equation)) <= 1/3 e.g.

	if sign_error_set = {} and wrong_factor_term_set = {} and
	   nops(term_error_set) = 1  and nops(missing_term_set) <= 1
	then RETURN([WRONG_TERMS, term_error_set, correct_equation]); fi;

	# only missing terms

	if sign_error_set = {} and wrong_factor_term_set = {} and
	   term_error_set = {}  and missing_term_set <> {}
	then RETURN([MISSING_TERMS, missing_term_set, correct_equation]); fi;

od;

#=== nothing specific identified, give up

RETURN([NOT_IDENTIFIED,{},0=0]); # 0=0 is a dummy here

end:

#============================================================================
#=====================  Check system of equations for correctness    ========
#============================================================================

G_check_system:=proc(system1,system2,var_set)

#=== system1 and system are assumed to be sets of linear equations,
#=== where system2 is assumed to be the correct system and system1 is checked.
#=== var_set is assumed to be the set of variables for which a system was solved
#=== Returns list [result,set_of_equations], which is [0,{}] if the systems are
#=== equivalent, and [1,{wrong equations}] if they are not equivalent.

local equation, solution1, correct_solution, extended_sol, wrong_equations;

wrong_equations:={}; # Initialize

solution1:=solve(system1, var_set);
if solution1 = NULL then solution1:={}; fi; # if solve can't solve
correct_solution:=solve(system2,var_set);
if correct_solution = NULL then correct_solution:={};fi; # if solve can't solve

if G_compare_solution_sets(solution1, correct_solution,var_set)=0

then RETURN([0,{}]);
else 
     for equation in system1 do
	 extended_sol:=solve(system2 union {equation}, var_set);
	 if extended_sol = NULL then extended_sol:={}; fi;  # if solve can't solve
	 if G_compare_solution_sets(extended_sol, correct_solution, var_set)=1 then
	 wrong_equations:=wrong_equations union {equation}; fi;
     od;
fi;
RETURN([1,wrong_equations]);
end:


#============================================================================
#=====================  Provide missing equations in system          ========
#============================================================================


G_provide_missing:=proc(system1, system2, var_set, number)

#=== system1 is assumed to be sub-system of system2 which is the complete system
#=== var_set is assumed to be the set of variables for which system2 is solved
#=== number is optional: number of missing equations to provide; can be used
#=== to find missing equations successively with a chance to go on without help
#=== Returns list [1,{}] if system1 is not a sub-system (i.e. dependent) of system2,
#=== otherwise [0,{missing equations}] (possibly restricted to a number of eq.)

local equation, missing_set, augmented_system, number_of_eq;
missing_set:={};
if nargs=4 then number_of_eq:=number; else number_of_eq:=infinity; fi;


# check first whether system1 is compatible with the "whole" system2

if G_check_system(system1 union system2,system2,var_set)[1] = 1 then 
	RETURN([1,{}]);   fi;



# add successively equations from system2 to system1 if they are independent from
# the system including formerly added equations

for equation in system2 do
	if nops(missing_set) = number_of_eq then break;fi;
	augmented_system:=system1 union missing_set;
	if G_check_system(augmented_system union {equation},augmented_system,var_set)[1]=1
   	then missing_set:=missing_set union {equation}; fi;
od;

# note that if the equations in system2 are ordered differently one will probably
# get different equations added

[0,missing_set];



end:


#============================================================================
#=====================  Extract symbol set from any expression       ========
#============================================================================


G_extract_symbol_set:=proc(any_expression)

# any_expression can be any expression, e.g. a system of equations

local i,symbol_set;
symbol_set:={};

if whattype(any_expression)=symbol or whattype(any_expression)=string then symbol_set:={any_expression};
elif type(any_expression,complex)=false then 
     for i from 1 to nops(any_expression) do
 	symbol_set:=symbol_set union G_extract_symbol_set(op(i,any_expression));
     od;
fi;
symbol_set;
end:


#============================================================================
#=====================  compare solution sets	                     ========
#============================================================================

G_compare_solution_sets:=proc(set1,set2, var_set)

#=== set1 and set2 are assumed to be solution set output from the solve procedure
#=== WARNING: solve might produce a NULL sequence in which case the sets must be
#=== set to {}.
#=== var_set is assumed to be the set of variables for which a system was solved

local variable, set_member, solution1, solution2;

for variable in var_set do

	for set_member in set1 do
		if variable=op(1,set_member) then
		   solution1:=op(2,set_member); break;fi;
	od;
	for set_member in set2 do
		if variable=op(1,set_member) then
		   solution2:=op(2,set_member); break;fi;
	od;

	if simplify(solution1 - solution2) <> 0 
	  then RETURN(1); fi;   # sets are not equal
od;

RETURN(0);   # sets are equal

end:



#============================================================================
#=====================  compare symbol sets	                     ========
#============================================================================

G_compare_symbol_sets:=proc(system1,system2)

#=== system1 is a set of equations or just one equation, system2 is a set of
#=== equations which are assumed to be the correct ones. The procedure returns
#=== a list with two sets of symbols: the first contains wrong symbols in system2,
#=== the second contains missing symbols

local set1,set2,result_list,i;

result_list:=array(1..2);
set1:=G_extract_symbol_set(system1);
set2:=G_extract_symbol_set(system2);

result_list[1]:=set1 minus set2; # should be empty set if system1 contains correct
				# symbols
result_list[2]:=set2 minus set1; # contains those symbols missing in set1

result_list;

end:

#============================================================================
#=====================  compare equations	                     ========
#============================================================================

G_compare_equations:=proc(equation1,equation2,var_set)

#=== check equation1 against equation2. Output a list containing the
#=== following: equivalence_flag: zero if equations are equivalent, 1 otherwise
#===            sign_error_set: set of terms with wrong sign in equation1
#===		wrong_factor_term_set: set of terms with wrong factor other than
#===				-1 in equation1.
#===		wrong_factor_list: list of wrong factors other than -1, can be 
#===				checked for equal entries.
#===            term_error_set: set of terms in equation1 which do not show
#===				up in equation2, even a different factor is 
#===				allowed.
#===		missing_term_set: set of term in equation2 which do not show up
#===				in equation1 or in the sign_error_set or in the
#===				wrong_factor_term_set, i.e. which are "completely"
#===				missing

local equiv_flag, new_eq1, new_eq2,i, set1, set2, setmember1, setmember2, counter,n,
	sign_error_set, wrong_factor_term_set, wrong_factor_list, term_error_set,
	missing_term_set, correct_factor_term_set, factorexpr, wrong_factor_seq;

#=== Initialization

sign_error_set:={};
wrong_factor_term_set:={};
wrong_factor_list:=[];
term_error_set:={};
missing_term_set:={};
correct_factor_term_set:={};
wrong_factor_seq:=NULL;

#=== check for equivalence

if G_check_system({equation1},{equation2},var_set)[1]=0 then 
		RETURN([0,{},{},[],{},{}]);fi;
		# if equivalent then return equiv_flag 0
#=== otherwise

equiv_flag:=1;

#=== make sure "everything" is on the left side of the equation

new_eq1:=simplify(lhs(equation1)-rhs(equation1))=0;
new_eq2:=simplify(lhs(equation2)-rhs(equation2))=0;

set1:={seq(op(i,lhs(new_eq1)),i=1..nops(lhs(new_eq1)))};
set2:={seq(op(i,lhs(new_eq2)),i=1..nops(lhs(new_eq2)))};

#=== check for terms with erroneous sign and other factor as well
#=== as "completely" wrong terms

n:=nops(set2);

for setmember1 in set1 do
	if member(setmember1,set2)=true then 0;
	elif member(-setmember1,set2)=true
		then sign_error_set:=sign_error_set union {setmember1};
	else counter:=0;
		for setmember2 in set2 do
		  counter:=counter+1;
		  factorexpr:=simplify(setmember1/setmember2);
		  if type(factorexpr,complex)=true
		     then wrong_factor_term_set:=
				wrong_factor_term_set union {setmember1};
		     wrong_factor_seq:=wrong_factor_seq, factorexpr;
		     correct_factor_term_set:=
				correct_factor_term_set union {setmember2};
		     break;
		  else if counter=n then 
			term_error_set:=term_error_set union {setmember1};fi;
		  fi;
		od;
	fi;
od;

wrong_factor_list:=[wrong_factor_seq];

#=== check for completely missing terms (not just sign errors)

missing_term_set:=set2 minus (set1 union 
			{seq(-sign_error_set[i],i=1..nops(sign_error_set))} 
			union correct_factor_term_set);

#=== now return the whole stuff

RETURN([equiv_flag, sign_error_set, wrong_factor_term_set, wrong_factor_list,
	term_error_set, missing_term_set]);

end:


